/*
  copyright (gpl)   19 sep 2012.
  Thorfinn Sand Korneliussen thorfinn@binf.ku.dk
  Anders albrechtsen albrecht@binf.ku.dk 
  0.12 more object oriented version of 0.03
  0.2 implemented bambi the native bamreader and GL estimator
  0.3 first real public, cleancing of program
  0.4 much is now better concerning screen output
*/

#include<iostream>//for printing time

#include<cstring> //for cstring functions
#include<cstdlib> //for exit()
#include<cstdio> //for fprintf
#include <signal.h>//for catching ctrl+c, allow threads to finish

#include "shared.h"
#include "multiReader.h"

#define VERSION 0.4471
#define ARGS ".arg"

#include "parseArgs_bambi.h"
#include "bammer_main.h"

extern std::vector <char *> dumpedFiles;//a vector of outputfilenames from shared.cpp

int SIG_COND =1;//if we catch signal then quit program nicely
int VERBOSE =1;
extern pthread_mutex_t mUpPile_mutex;//mutex for not deleting mUppile data untill end of data

void printTime(FILE *fp){
 time_t rawtime;
  struct tm * timeinfo;

  time ( &rawtime );
  timeinfo = localtime ( &rawtime );
  fprintf (fp, "%s", asctime (timeinfo) );
}


//this function is called from within the bamreader
void callBack_bambi(fcb *fff){

  if(fff==NULL)
    selector(NULL);//<-send NULL which acts as a killswitch
  else{

    fflush(stderr);
    funkyPars *fp = allocFunkyPars();
    fp->for_callback=fff;
    selector(fp);
  }
}



void handler(int s) {
  if(VERBOSE)
    fprintf(stderr,"Caught SIGNAL: %d will try to exit nicely (no more threads are created, we will wait for the current threads to finish)\n",s);
  VERBOSE=0;
  SIG_COND=0;
  pthread_mutex_unlock(&mUpPile_mutex);
}

 //we are threading so we want make a nice signal handler for ctrl+c
void catchkill(){
  struct sigaction sa;
  sigemptyset (&sa.sa_mask);
  sa.sa_flags = 0;
  sa.sa_handler = handler;
  sigaction(SIGPIPE, &sa, 0);
  sigaction(SIGINT, &sa, 0);  

}

//print nice info
void printProgInfo(FILE *fp){

  fprintf(fp,"\t->Please use the manual at www.popgen.dk/wiki\n\n");
  fprintf(fp,"Overview of methods:\n");
  fprintf(fp,"\t-GL\t\testimate genotype likelihoods\n");
  fprintf(fp,"\t-doAsso\t\tperform association study\n");
  fprintf(fp,"\t-doMaf\t\testimate allele frequencies\n");
  fprintf(fp,"\t-doMajorMinor\tinfer the major and minor allele\n");
  fprintf(fp,"\t-SNP\t\tdetect polymorpic sites\n");
  fprintf(fp,"\t-doError\testimate the type specific error rates\n");
  fprintf(fp,"\t-doCount\tcount the reads persite\n");
  fprintf(fp,"\t-ref/-ans\tread referenec or ancestral genome\n");
  fprintf(fp,"\t-doHWE\tEst inbreedning per site\n");
  fprintf(fp,"\t-doGeno\tcall genotypes\n");
  fprintf(fp,"\t-\trealSFS\n");
  fprintf(fp,"\t-covar\tapproximate the covariance matrix\n");
  fprintf(fp,"\tmany others\n\n");
  
}




void parseArgStruct(argStruct *arguments){
  fflush(stderr);
  for(int i=1;i<arguments->argc;i++){
    if(arguments->usedArgs[i]==0){
      fprintf(stderr,"%d argument \t%s is unknown will exit\n",i,arguments->argv[i]);
      fflush(stderr);
      exit(0);
    }
  }

}




 void getNind(const char *name,argStruct *arguments){
  if(!fexists(name)){
    fprintf(stderr,"[%s]\t-> Problems opening file: %s\n",__FUNCTION__,name);
    return;
     exit(0);
  }
  int nInd = -1;
  nInd = angsd::getArg("-nInd",nInd,arguments);
  //  fprintf(stderr,"ASDF nInd=%d\n",nInd);

  std::vector<char*> getFilenames2(const char * name,int nInd);
  std::vector<char*> delme = getFilenames2(name,nInd);
  arguments->nInd=delme.size();
  fprintf(stderr,"\t->nInd=%d in %s\n",arguments->nInd,name);
  
  for(uint i=0;i<delme.size();i++)
    free(delme[i]);
 }


argStruct *setArgStruct(int argc,char **argv) { 
  argStruct *arguments = new argStruct;
  arguments->argc=argc;
  arguments->argv=argv;
  arguments->usedArgs= new int[argc];
  for(int i=0;i<argc;i++)
    arguments->usedArgs[i]=0;
  
  arguments->inputtype=-1;
  int argPos = 1;
  while(argPos <argc) { 
    if (strcmp(argv[argPos],"-samglf")==0){
      arguments->inputtype=1;
    } else if (strcmp(argv[argPos],"-samglfclean")==0){
      arguments->inputtype=2;
    } else if (strcmp(argv[argPos],"-sim1")==0){
      //      fprintf(stderr,"settting sim1\n");
       arguments->inputtype=4;
       arguments->nInd=angsd::getArg("-nInd",arguments->nInd,arguments);
     }
     else if (strcmp(argv[argPos],"-beagle")==0){
       arguments->inputtype=5;
       arguments->usedArgs[argPos]=1;
       arguments->usedArgs[argPos+1]=1;
     }
     else if (strcmp(argv[argPos],"-soap")==0){
       arguments->inputtype=0;
       arguments->usedArgs[argPos]=1;
       if(argc==2)
	 break;
     } else if (strcmp(argv[argPos],"-tglf")==0){
       arguments->inputtype=3;
       arguments->usedArgs[argPos]=1;
       arguments->usedArgs[argPos+1]=1;
     }     else if(strcmp(argv[argPos],"-out")==0){
       arguments->outfiles = argv[argPos+1];
       arguments->usedArgs[argPos]=1;
       arguments->usedArgs[argPos+1]=1;
     }
     else if(strcmp(argv[argPos],"-bam")==0||strcmp(argv[argPos],"-i")==0 ){
	 if(strcmp(argv[argPos],"-bam")==0)
	   getNind(argv[argPos+1],arguments);
	 else
	   arguments->nInd = 1;
	 arguments->inputtype=7;
	 arguments->usedArgs[argPos]=1;
	 arguments->usedArgs[argPos+1]=1;
       }
     argPos+=2;

   }
  //  fprintf(stderr,"inputtype=%d\n",arguments->inputtype);
  if(arguments->inputtype==-1 && argc!=2){
    fprintf(stderr,"\t-> Error: supply inputfiles:[-i afile.bam -bam bamlist, -samglf, -beagle, -samglfclean, -tglf]\n");
    exit(0);
  }

   return arguments;
 }




 int main(int argc, char** argv){

   //print time
   clock_t t=clock();
   time_t t2=time(NULL);

   //intialize our signal handler for ctrl+c
   catchkill();

   //print arguments supplied
   fprintf(stderr,"Command:\n");
   for(int i=0;i<argc;i++)
     fprintf(stderr,"%s ",argv[i]);
   fprintf(stderr,"\n\t-> angsd version: %.3f\t build(%s %s)\n",VERSION,__DATE__,__TIME__); 

   //no arguments supplied -> print info
   if(argc==1){//if haven't been supplied with arguments, load default,print, and exit
     printProgInfo(stderr);
     return 0;
   }

   argStruct *arguments = NULL;
   arguments = setArgStruct(argc,argv);
   arguments->argumentFile=stderr;
   //catch the informative case where we want to view options associated with a command
   if(argc==2){
     //if we dont use bamfiles
     multiReader mr(arguments->inputtype,arguments);
     FILE *tempFile;
     fprintf(stderr,"\t->info\n");
     args *bamPars=bamPars=getArgsBambi(arguments);
     init(arguments);//program dies here after printing info
   }
   //initialize argStruct containing needed stuff.


   

   
   //check that user supplied an output prefix for files
   if(arguments->outfiles ==NULL){
     fprintf(stderr,"\t-> Must supply -out\n");
     return 0;
   }

   //print arguments into logfile
   FILE *argFp = openFile(arguments->outfiles,ARGS);
   arguments->argumentFile=argFp;
   multiReader mr(arguments->inputtype,arguments);

   args *bamPars =NULL;
   if(arguments->inputtype==7){
     if(!(bamPars=getArgsBambi(arguments))){
       bamInfo(stderr);
	return 0;
     }
   }
   
   for(int i=0;i<argc;i++)
     fprintf(argFp,"%s ",argv[i]);
   fprintf(argFp,"\n\n");
   fprintf(argFp,"\n\t-> angsd version: %.3f\t build(%s %s)\n",VERSION,__DATE__,__TIME__); 
   fprintf(argFp,"\t->");
   printTime(argFp);  


   init(arguments);
   parseArgStruct(arguments);//validate that all arguments has indeed been parsed in the instantiation 'init' above


   //Below is main loop which will run until nomore data
   if(arguments->inputtype!=7){
     fprintf(stderr,"\t-> multireader\n");
    while(SIG_COND) {

      funkyPars *tmp = mr.fetch(); //YES MISTER FETCH
      if(tmp==NULL)
	break;  
      selector(tmp);
    }
  }else 
    bammer_main(bamPars);
  
   //now we have no more data lets wait and cleanup

  fprintf(stderr,"\n\t-> Done reading data waiting for calculations to finish\n");

  destroy();

  
  //printout the filenames generated
  //  fprintf(stderr,"[%s] bang BANG  you are done\n",__FILE__);
  fprintf(stderr,"\t->Output filenames:\n");
  for(int i=0;i<(int)dumpedFiles.size();i++){
    fprintf(stderr,"\t\t->\"%s\"\n",dumpedFiles[i]);
    fprintf(argFp,"\t\t->\"%s\"\n",dumpedFiles[i]);
    free(dumpedFiles[i]);
  }
  fprintf(stderr,"\n");
  fprintf(argFp,"\n");
  fprintf(stderr,"\t");
  printTime(stderr);
  
  delete []   arguments->usedArgs;
  delete arguments;
  //print out nice messages

  fprintf(stderr, "\t[ALL done] cpu-time used =  %.2f sec\n", (float)(clock() - t) / CLOCKS_PER_SEC);
  fprintf(stderr, "\t[ALL done] walltime used =  %.2f sec\n", (float)(time(NULL) - t2));  
  fprintf(argFp, "\t[ALL done] cpu-time used =  %.2f sec\n", (float)(clock() - t) / CLOCKS_PER_SEC);
  fprintf(argFp, "\t[ALL done] walltime used =  %.2f sec\n", (float)(time(NULL) - t2));  
  fclose(argFp);
  
  return 0;
}
