/*
  assumes zero index positions

  should remove mutexs, and make function thread safe. dragon daemon
 */



#include "analysisFunction.h"
#include "shared.h"
#include "faidx.h"
#include "general.h"
#include "pthread.h"

pthread_mutex_t getFasta_mut1 = PTHREAD_MUTEX_INITIALIZER;
pthread_mutex_t getFasta_mut2 = PTHREAD_MUTEX_INITIALIZER;



class getFasta : public general{
private:
  char *ancName;
  char *refName;
  char *getRef(const loci *locs,int lens,int doCleanup,faidx_t *fai);
  char *getAnc(const loci *locs,int lens,int doCleanup,faidx_t *fai);
  faidx_t *faiRef;
  faidx_t *faiAnc;
public:
  //  char *getRef2(const loci *locs,int lens,int doCleanup);
  // char *getRef1(const loci *locs,int lens,int doCleanup);
  getFasta(argStruct *arguments);
  ~getFasta();
  void printArg(FILE *argFile);
  void getOptions(argStruct *arguments);
  void run(funkyPars  *pars);//plugs in reference and ancestral
  void print(funkyPars *pars);
  void clean(funkyPars *pars);

};




void getFasta::printArg(FILE *argFile){
  fprintf(argFile,"%s:\n\n",__FILE__);
  fprintf(argFile,"-ref\t%s\n",refName);
  fprintf(argFile,"-anc\t%s\n",ancName);
  fprintf(argFile,"\n");

}

getFasta::~getFasta(){
  if(refName!=NULL)
    getRef(NULL,0,1,faiRef);//<-last indicate a cleanup
  if(ancName!=NULL)
    getAnc(NULL,0,1,faiAnc);//<-last indicate a cleanup

  if(refName!=NULL&&faiRef!=NULL)
    fai_destroy(faiRef);
  if(ancName!=NULL&&faiAnc!=NULL)
    fai_destroy(faiAnc);

}


void getFasta::getOptions(argStruct *arguments){
  //default
  refName = NULL;
  ancName = NULL;
  //from command line
  ancName = angsd::getArg("-anc",ancName,arguments);
  refName = angsd::getArg("-ref",refName,arguments);
  //  fprintf(stderr,"refName=%s\tancName=%s\n",refName,ancName);


  if(ancName!=NULL)
    if(strcmp(ancName,"-999")==0){
      ancName=NULL;
      printArg(stdout);
      exit(0);
 
    }
  if(refName!=NULL)
    if(strcmp(refName,"-999")==0){
      refName=NULL;
      printArg(stdout);
      exit(0);
 
    }
  if(ancName==NULL|refName==NULL)
    return;
  printArg(arguments->argumentFile);

}

 getFasta::getFasta(argStruct *arguments){

   //   fprintf(stderr,"index=%d tot_index=%d\n",index,tot_index);
   getOptions(arguments);

  if(refName!=NULL)
    faiRef = fai_load(refName);
  if(ancName!=NULL)
    faiAnc = fai_load(ancName);
}

void getFasta::run(funkyPars *pars){
  if(pars->numSites==0)
    return;
  if(refName!=NULL)
    pars->ref = getRef(pars->sites,pars->numSites,0,faiRef);
  if(ancName!=NULL)
    pars->anc = getAnc(pars->sites,pars->numSites,0,faiAnc);
  
}


char *getFasta::getRef(const loci *locs,int lens,int doCleanup,faidx_t *fai) {
  pthread_mutex_lock(&getFasta_mut1);
  static char *chrLast =NULL;
  static char *ref = NULL;
  static int ref_len = -1;
  static int missingChrInFai =0;

  if(doCleanup){
    free(chrLast);
    free(ref);
    pthread_mutex_unlock(&getFasta_mut1);
    return NULL;
  }

  //  fprintf(stderr,"\t-> Reading fasta ref/anc [%s] fname=%s\n",__FUNCTION__,fasta);

  char *ret=new char[lens];


  static int count=0;
  //only runs if no previsou chromsome
  if(chrLast==NULL||(strcmp(chrLast,locs[0].chromo)!=0)){
    missingChrInFai =0;
    //    fprintf(stdout,"chrLast herehe:%d\n",count++);
    chrLast =strdup(locs[0].chromo);//THORFINN leak.. should fix.
    free(ref);
    ref = faidx_fetch_seq(fai, chrLast, 0, 0x7fffffff, &ref_len);
    //    fprintf(stderr,"%s  1) reflen=%d\n",__FILE__,ref_len);
    if(ref==NULL){
      fprintf(stderr,"[%s] Error loading reference alleles 1\n",__FUNCTION__);
      //      exit(0);
      missingChrInFai =1;
    }
  }
  for(int i=0;i<lens;i++){
    if(strcmp(locs[i].chromo,chrLast)!=0){
      missingChrInFai =0;
      free(ref);
      ref = faidx_fetch_seq(fai, chrLast, 0, 0x7fffffff, &ref_len);
      //      fprintf(stderr,"%s  2) reflen=%d\n",__FILE__,ref_len);
      if(ref==NULL){
	fprintf(stderr,"[%s] Error loading reference alleles 2\n",__FUNCTION__);
	missingChrInFai =1;
	//	exit(0);
      }
    }
    //    fprintf(stderr,"%d\n",ref_len);
    //ret[i] = refToInt[ref[locs[i].position-1]];//offseting by one, to make in work like samtools
    if(missingChrInFai==0){
      if(locs[i].position>ref_len){
	fprintf(stderr,"Warning position is after fasta :%s/%s pos=%d ref_len=%d\n",chrLast,locs[i].chromo,locs[i].position,ref_len);
	if(locs[i].position+200>ref_len){
	  fprintf(stderr,"Trying to access fasta efter end of chromsome+200:%s/%s pos=%d ref_len=%d\n",chrLast,locs[i].chromo,locs[i].position,ref_len);
	  fflush(stderr);
	  exit(0);
	}
	ret[i]=4;
	continue;
      }
    }
    if(missingChrInFai)
      ret[i] = 4;
    else
      ret[i] = refToInt[ref[locs[i].position]];//offseting by one, to make in work like samtools

  }
  pthread_mutex_unlock(&getFasta_mut1);

  return ret;

}

char *getFasta::getAnc(const loci *locs,int lens,int doCleanup,faidx_t *fai){
  pthread_mutex_lock(&getFasta_mut2);
  static char *chrLast =NULL;
  static char *ref = NULL;

  static int ref_len;
  static int missingChrInFai =0;
  if(doCleanup){
    free(chrLast);
    free(ref);
    pthread_mutex_unlock(&getFasta_mut2);
    return NULL;
  }

  char *ret=new char[lens];

  if(chrLast==NULL||(strcmp(chrLast,locs[0].chromo)!=0)){
    missingChrInFai =0;
    chrLast =strdup(locs[0].chromo);//THORFINN leak.. should fix.
    free(ref);
    ref = faidx_fetch_seq(fai, chrLast, 0, 0x7fffffff, &ref_len);
    //    fprintf(stderr,"fetching");
    if(ref==NULL){
      fprintf(stderr,"[%s] Error loading anc/reference alleles chr=%s\n",__FUNCTION__,locs[0].chromo);
      //      exit(0);
      missingChrInFai =1;
    }

  }
  for(int i=0;i<lens;i++){
    if(strcmp(locs[i].chromo,chrLast)!=0){
      missingChrInFai =0;
      free(ref);
      chrLast=locs[i].chromo;
      ref=NULL;
      ref = faidx_fetch_seq(fai, chrLast, 0, 0x7fffffff, &ref_len);
      if(ref==NULL){
	fprintf(stderr,"[%s] Error loading anc/reference alleles\n",__FUNCTION__);
	//	exit(0);
	missingChrInFai =1;
      }
    }
   if(missingChrInFai==0){
      if(locs[i].position>ref_len){
	fprintf(stderr,"Warning position is after fasta :%s/%s pos=%d ref_len=%d\n",chrLast,locs[i].chromo,locs[i].position,ref_len);
	if(locs[i].position+200>ref_len){
	  fprintf(stderr,"Trying to access fasta efter end of chromsome+200:%s/%s pos=%d ref_len=%d\n",chrLast,locs[i].chromo,locs[i].position,ref_len);
	  fflush(stderr);
	  exit(0);
	}
	ret[i]=4;
	continue;
      }
    }
    if(missingChrInFai)
      ret[i] = 4;
    else
      ret[i] = refToInt[ref[locs[i].position]];//offseting by one, to make in work like samtools

  }
  pthread_mutex_unlock(&getFasta_mut2);
  return ret;
}

void getFasta::print(funkyPars *f){


}
void getFasta::clean(funkyPars *f){


}
